@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2011 Emil Mikulic.
+ * copyright (c) 2001-2012 Emil Mikulic.
  *
  * cap.h: interface to libpcap.
  */
@@ -10,12 +10,14 @@
 
 extern unsigned int cap_pkts_recv, cap_pkts_drop;
 
-void cap_init(const char *device, const char *filter, int promisc);
+void cap_add_ifname(const char *ifname); /* call one or more times */
+void cap_add_filter(const char *filter); /* call zero or more times */
+void cap_start(const int promisc);
 void cap_fd_set(fd_set *read_set, int *max_fd,
    struct timeval *timeout, int *need_timeout);
 void cap_poll(fd_set *read_set);
 void cap_stop(void);
 
-void cap_from_file(const char *capfile, const char *filter);
+void cap_from_file(const char *capfile);
 
 /* vim:set ts=3 sw=3 tw=78 expandtab: */
