@@ -49,6 +49,11 @@ struct {								\
 
 #define	STAILQ_FIRST(head)	((head)->stqh_first)
 
+#define STAILQ_FOREACH(var, head, field)                                \
+        for((var) = STAILQ_FIRST((head));                               \
+           (var);                                                       \
+           (var) = STAILQ_NEXT((var), field))
+
 #define	STAILQ_NEXT(elm, field)	((elm)->field.stqe_next)
 
 #ifdef STAILQ_INSERT_TAIL
